from __future__ import annotations
from typing import TYPE_CHECKING, Type

import re

if TYPE_CHECKING:
    from volga.types import supportsDeser

from typing import get_type_hints
from volga.fields import Bool, Float, Int, Null, Str


from volga.format import Format

RE_FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL

# parse string after first '"' and terminate by '"'
STRING_RE = re.compile(r'(.*?)(")', RE_FLAGS)

# to be used for parsing JSON numbers
NUMBER_RE = re.compile(
    r"(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?",
    RE_FLAGS,
)


class JSON(Format):
    def __init__(self, input: str) -> None:
        self.s: str = input
        self.idx: int = 0

    def parse_bool(self) -> bool:

        idx = self.idx
        curr_char = self.s[idx]

        if curr_char == "t" and self.s[idx : idx + 4] == "true":
            # move to 1 pass the last letter of true
            self.idx += 5
            return True
        elif curr_char == "f" and self.s[idx : idx + 5] == "false":
            self.idx += 6
            return False
        else:
            raise RuntimeError("Expected bool")

    def parse_float(self) -> float:

        match = NUMBER_RE.match(self.s, self.idx)

        if match:
            integer, frac, exp = match.groups()

            n = float(integer + (frac or "") + (exp or ""))
            self.idx = match.end()
            return n
        else:
            raise RuntimeError("Expected float")

    def parse_int(self) -> int:

        match = NUMBER_RE.match(self.s, self.idx)

        if match:
            integer, frac, exp = match.groups()
            if frac or exp:
                raise RuntimeError("Expected integer value.")
            else:
                n = int(integer)
            self.idx = match.end()
            return n
        else:
            raise RuntimeError("Expected int")

    def parse_str(self) -> str:

        if self.s[self.idx] != '"':
            raise RuntimeError("Expected String")

        # skip past first "
        chunk = STRING_RE.match(self.s, self.idx + 1)

        if chunk is None:
            raise RuntimeError("Unterminated string.")

        content, _ = chunk.groups()
        self.idx = chunk.end()
        return content

    def parse_none(self) -> None:
        idx = self.idx
        curr_char = self.s[idx]

        if curr_char == "n" and self.s[idx : idx + 4] == "null":
            self.idx += 5
            return None
        else:
            raise RuntimeError("Expected null")

    def __deserialize_str__(self, constructor: Type[supportsDeser]) -> supportsDeser:
        return constructor.__from_str__(self.parse_str())

    def __deserialize_dict__(self, constructor: Type[supportsDeser]) -> supportsDeser:

        res = {}

        if self.s[self.idx] != "{":
            raise RuntimeError("Expected dict")

        # for each attribute in the schema

        attrs = get_type_hints(constructor, include_extras=True)
        for key in attrs:

            # skip past first { and subsequent ,
            chunk = STRING_RE.match(self.s, self.idx + 2)

            if chunk is None:
                raise RuntimeError("Expected key string.")

            parsed_key, _ = chunk.groups()
            self.idx = chunk.end()

            assert parsed_key == key

            if self.s[self.idx] != ":":
                raise RuntimeError("Expected value for key")
            self.idx += 1

            # parse the value according to schema
            print(attrs[key])
            value = self.dispatch(attrs[key])

            res[Str(parsed_key)] = value

        return constructor.__from_dict__(res)

    def __deserialize_bool__(self, constructor: Type[supportsDeser]) -> supportsDeser:
        return constructor.__from_bool__(self.parse_bool())

    def __deserialize_int__(self, constructor: Type[supportsDeser]) -> supportsDeser:
        return constructor.__from_int__(self.parse_int())

    def __deserialize_float__(self, constructor: Type[supportsDeser]) -> supportsDeser:
        return constructor.__from_float__(self.parse_float())

    def __deserialize_none__(self, constructor: Type[supportsDeser]) -> supportsDeser:
        # consume null
        self.parse_none()
        return constructor.__from_none__(None)

    def dispatch(self, cls: Type[supportsDeser]) -> supportsDeser:
        if issubclass(cls, Bool):
            return cls(False).__deserialize__(self)
        elif issubclass(cls, Int):
            return cls(0).__deserialize__(self)
        elif issubclass(cls, Float):
            return cls(0.0).__deserialize__(self)
        elif issubclass(cls, Str):
            return cls("").__deserialize__(self)
        elif issubclass(cls, Null):
            return cls(None).__deserialize__(self)
        else:
            return cls({}).__deserialize__(self)


def deserialize(input: str, cls: Type[supportsDeser]) -> supportsDeser:

    format = JSON(input)

    # initialize empty instance to dispatch deserialize
    # TODO: investigate why protocol is not working with class method

    return format.dispatch(cls)



from __future__ import annotations

from abc import abstractmethod
from typing import Protocol, TYPE_CHECKING, Type, Any


if TYPE_CHECKING:
    from volga.format import Format


class supportsDeser(Protocol):
    """This protocol defines the entry point for objects to be deserializable"""

    @abstractmethod
    def __init__(self, value: Any) -> None:
        ...

    @abstractmethod
    def __deserialize__(self, format: Format) -> supportsDeser:
        raise NotImplementedError

    @classmethod
    def __from_bool__(cls: Type[supportsDeser], value: bool) -> supportsDeser:
        return cls(value)

    @classmethod
    def __from_int__(cls: Type[supportsDeser], value: int) -> supportsDeser:
        return cls(value)

    @classmethod
    def __from_float__(cls: Type[supportsDeser], value: float) -> supportsDeser:
        return cls(value)

    @classmethod
    def __from_str__(cls: Type[supportsDeser], value: str) -> supportsDeser:
        return cls(value)

    @classmethod
    def __from_none__(cls: Type[supportsDeser], value: Any) -> supportsDeser:
        return cls(value)

    @classmethod
    def __from_dict__(cls: Type[supportsDeser], d: dict[Any, Any]) -> supportsDeser:
        return cls(d)



from __future__ import annotations
from typing import Any

from volga.types import supportsDeser
from volga.format import Format


class Schema(supportsDeser):
    def __init__(self, d: dict[Any, Any]) -> None:
        for key in d:
            setattr(self, key.__str__(), d[key])

    def __str__(self) -> str:
        return str(vars(self))

    @classmethod
    def __deserialize__(cls, format: Format) -> supportsDeser:
        """deserialize schemas from dictionaries"""
        return format.__deserialize_dict__(cls)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['volga']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'volga',
    'version': '0.2.0',
    'description': 'A python framework for deserialization',
    'long_description': '# volga: flexible object deserialization\n\n[![Build Status]][build] [![Azure DevOps coverage]][Azure coverage url] [![license]][license-file] [![release]][releases] [![python-version]][pypi]\n[![Docs](https://img.shields.io/readthedocs/volga.svg)](https://volga.readthedocs.io)\n\n[Build Status]: https://dev.azure.com/yefrigaitan/volga/_apis/build/status/yefrig.volga?branchName=main\n[build]: https://dev.azure.com/yefrigaitan/volga/_build/latest?definitionId=9&branchName=main\n\n[Azure DevOps coverage]: https://img.shields.io/azure-devops/coverage/yefrigaitan/volga/9\n[Azure coverage url]: https://dev.azure.com/yefrigaitan/volga/_build/latest?definitionId=9&branchName=main\n\n[license]: https://img.shields.io/github/license/yefrig/volga\n[license-file]: https://github.com/yefrig/volga/blob/main/LICENSE\n\n[release]: https://img.shields.io/github/v/release/yefrig/volga?include_prereleases&sort=semver\n[releases]: https://github.com/yefrig/volga/releases\n\n[python-version]: https://img.shields.io/pypi/pyversions/volga\n[pypi]: https://pypi.org/project/volga/\n\n## What is it?\n**volga** provides fast, extensible, and expressive APIs\nto deserialize any python data structure from any supported data format\n(such as JSON and *eventually* YAML and more). Volga allows full customization of the deserialization \nbehavior of your data structures resulting in schema-tized, validated, type-checked \nobjects.\n\n```python3\nimport volga\n\n# Define your model\nclass User(volga.Schema):\n    name: volga.fields.Str\n    age: volga.fields.Int\n    verified: volga.fields.Bool\n  \njson_data = \'{"name":"bob","age":20,"verified":true}\'\n\nbob = volga.json.deserialize(json_data, User)\n\nassert isinstance(bob, User)\n\nprint(bob) # prints object User(name=\'bob\', age=20, verified=True)\n```\n\n## Main Features\n\n\n## Documentation\n\nFull documentation will soon be available on https://volga.readthedocs.io/en/latest/\n\n\n## Where to get it\nThe source code is currently hosted on GitHub at:\nhttps://github.com/yefrig/volga\n\nBinary installers for the latest released version are available at the [Python\npackage index](https://pypi.org/project/volga).\n\n```sh\npip install volga\n```\n\n## Main Contributors\n\n- Yefri Gaitan [@yefrig](https://github.com/yefrig)\n\n- Ecenaz (Jen) Ozmen [@eozmen410](https://github.com/eozmen410)\n',
    'author': 'Yefri Gaitan',
    'author_email': 'yefrigaitan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yefrig/volga',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from abc import abstractmethod
from typing import Protocol, TYPE_CHECKING, Type

if TYPE_CHECKING:
    from volga.types import supportsDeser


class Format(Protocol):
    """This protocol defines the methods data formats should support
        if they want to interact with volga's types.
        if they want to interact with volga's Field types.

    :raises NotImplementedError: Classes that subclass this protocol
        should not called any of these functions using super().
    """

    @abstractmethod
    def __deserialize_bool__(self, cls: Type[supportsDeser]) -> supportsDeser:
        raise NotImplementedError

    @abstractmethod
    def __deserialize_int__(self, cls: Type[supportsDeser]) -> supportsDeser:
        raise NotImplementedError

    @abstractmethod
    def __deserialize_float__(self, cls: Type[supportsDeser]) -> supportsDeser:
        raise NotImplementedError

    @abstractmethod
    def __deserialize_str__(self, cls: Type[supportsDeser]) -> supportsDeser:
        raise NotImplementedError

    @abstractmethod
    def __deserialize_dict__(self, cls: Type[supportsDeser]) -> supportsDeser:
        raise NotImplementedError

    @abstractmethod
    def __deserialize_none__(self, cls: Type[supportsDeser]) -> supportsDeser:
        raise NotImplementedError



from __future__ import annotations
from abc import ABC
from typing import Any

from volga.types import supportsDeser
from volga.format import Format


class Field(ABC):
    """
    Provides methods for a data format to construct this
    field from any of the supported volga data types.
    """


class Bool(int, Field, supportsDeser):
    def __init__(self, value: Any) -> None:
        super().__init__(value)

    def __repr__(self) -> str:
        return bool.__repr__(bool(self))

    def __deserialize__(self, format: Format) -> supportsDeser:
        return format.__deserialize_bool__(self.__class__)


class Int(int, Field, supportsDeser):
    def __init__(self, value: Any) -> None:
        super().__init__(value)

    def __deserialize__(self, format: Format) -> supportsDeser:
        return format.__deserialize_int__(self.__class__)


class Float(float, Field, supportsDeser):
    def __init__(self, value: Any) -> None:
        super().__init__(value)

    def __deserialize__(self, format: Format) -> supportsDeser:
        return format.__deserialize_float__(self.__class__)


class Str(str, Field, supportsDeser):
    def __init__(self, value: Any) -> None:
        super().__init__(value)

    def __deserialize__(self, format: Format) -> supportsDeser:
        return format.__deserialize_str__(self.__class__)


class Null(Field, supportsDeser):

    """ Null objects always and reliably "do nothing." """

    def __init__(self, value: Any) -> None:
        return None

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        return None

    def __repr__(self) -> str:
        return "None"

    def __nonzero__(self) -> int:
        return 0

    def __deserialize__(self, format: Format) -> supportsDeser:
        return format.__deserialize_none__(self.__class__)

    def __eq__(self, other: Any):
        return other is None


class Dict(dict[Any, Any], Field, supportsDeser):
    def __init__(self, value: Any) -> None:
        super().__init__(value)

    def __deserialize__(self, format: Format) -> supportsDeser:
        return format.__deserialize_dict__(self.__class__)
